#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "stack.h"

int32_t operators_length = 5;
char operators_symbol[] = {'+', '-', '*', '/', '^'};
int32_t operators_precedence[] = {1, 1, 2, 2, 3};

void print_stack(Stack *s) {
    for (int32_t i = 0; i < s->top + 1; i += 1) {
        printf("%s,", s->data[i]);
    }

    printf("\n");
}

int32_t get_operator_precedence(char operator) {
    for (int32_t i = 0; i < operators_length; i += 1) {
        if (operators_symbol[i] == operator) {
            return operators_precedence[i];
        }
    }

    return -1;
}

bool is_operator(char c) {
    for (int32_t i = 0; i < operators_length; i += 1) {
        if (operators_symbol[i] == c) {
            return true;
        }
    }

    return false;
}

char **infix_to_postfix(char *infix_expression[], int32_t infix_expression_length, int32_t *postfix_epxression_length) {
    Stack operators;
    stack_init(&operators, infix_expression_length);
    Stack output;
    stack_init(&output, infix_expression_length);

    for (int32_t i = 0; i < infix_expression_length; i += 1) {
        if (infix_expression[i][0] == '(') {
            stack_push(&operators, infix_expression[i]);
        } else if (infix_expression[i][0] == ')') {
            while (true) {
                char operator[100];
                stack_pop(&operators, operator);

                if (operator[0] == '(') {
                    break;
                }

                stack_push(&output, operator);
            }
        } else if (is_operator(infix_expression[i][0])) {
            if (stack_is_empty(&operators)) {
                stack_push(&operators, infix_expression[i]);
            } else {
                while (!stack_is_empty(&operators)) {
                    char operator[100];
                    stack_peek(&operators, operator);
                    int32_t precedence_1 = get_operator_precedence(operator[0]);
                    int32_t precedence_2 = get_operator_precedence(infix_expression[i][0]);

                    if (precedence_2 > precedence_1) {
                        break;
                    }

                    char tmp[100];
                    stack_pop(&operators, tmp);
                    stack_push(&output, tmp);
                }

                stack_push(&operators, infix_expression[i]);
            }
        } else {
            stack_push(&output, infix_expression[i]);
        }
    }

    while (!stack_is_empty(&operators)) {
        char tmp[100];
        stack_pop(&operators, tmp);
        stack_push(&output, tmp);
    }

    char **output_str = (char **)malloc(sizeof(char *) * (output.top + 1));

    for (int32_t i = 0; i < (output.top + 1); i += 1) {
        output_str[i] = (char *)malloc(sizeof(char) * 100);
        strcpy(output_str[i], output.data[i]);
    }

    *postfix_epxression_length = output.top + 1;

    stack_destroy(&operators);
    stack_destroy(&output);

    return output_str;
}

double parse_postfix_epxression(char *postfix_expression[], int32_t postfix_expression_length) {
    Stack output;
    stack_init(&output, postfix_expression_length);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        if (!is_operator(postfix_expression[i][0])) {
            stack_push(&output, postfix_expression[i]);
            continue;
        }

        char operand_1[100];
        char operand_2[100];
        stack_pop(&output, operand_2);
        stack_pop(&output, operand_1);
        double result;

        switch (postfix_expression[i][0]) {
            case '+':
                result = atof(operand_1) + atof(operand_2);
                break;
            case '-':
                result = atof(operand_1) - atof(operand_2);
                break;
            case '*':
                result = atof(operand_1) * atof(operand_2);
                break;
            case '/':
                result = atof(operand_1) / atof(operand_2);
                break;
            case '^':
                result = pow(atof(operand_1), atof(operand_2));
                break;
        }

        char result_str[100];
        sprintf(result_str, "%.15lf", result);
        stack_push(&output, result_str);
    }

    char result_str[100];
    stack_pop(&output, result_str);
    stack_destroy(&output);
    return atof(result_str);
}
