#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "calculator.h"

const double EPSILON = 1e-10;

void test_1() {
    // 7+2*3
    char *infix_expression[] = {"7", "+", "2", "*", "3"};
    int32_t infix_expression_length = 5;

    int32_t postfix_expression_length;
    char **postfix_expression = infix_to_postfix(infix_expression, infix_expression_length, &postfix_expression_length);

    double result = parse_postfix_epxression(postfix_expression, postfix_expression_length);
    assert(fabs(13 - result) <= EPSILON);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        free(postfix_expression[i]);
    }
    free(postfix_expression);
}

void test_2() {
    // 8-5-2
    char *infix_expression[] = {"8", "-", "5", "-", "2"};
    int32_t infix_expression_length = 5;

    int32_t postfix_expression_length;
    char **postfix_expression = infix_to_postfix(infix_expression, infix_expression_length, &postfix_expression_length);

    double result = parse_postfix_epxression(postfix_expression, postfix_expression_length);
    assert(fabs(1 - result) <= EPSILON);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        free(postfix_expression[i]);
    }
    free(postfix_expression);
}

void test_3() {
    // (7-5)/(2-(1-9)^(2^3))
    char *infix_expression[] = {"(", "7", "-", "5", ")", "/", "(", "2", "-", "(", "1", "-", "9", ")", "^", "(", "2", "^", "3", ")", ")"};
    int32_t infix_expression_length = 21;

    int32_t postfix_expression_length;
    char **postfix_expression = infix_to_postfix(infix_expression, infix_expression_length, &postfix_expression_length);

    double result = parse_postfix_epxression(postfix_expression, postfix_expression_length);
    assert(fabs(-0.00000011921 - result) <= EPSILON);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        free(postfix_expression[i]);
    }
    free(postfix_expression);
}

void test_4() {
    // 2/(7-3*2)
    char *infix_expression[] = {"2", "/", "(", "7", "-", "3", "*", "2", ")"};
    int32_t infix_expression_length = 9;

    int32_t postfix_expression_length;
    char **postfix_expression = infix_to_postfix(infix_expression, infix_expression_length, &postfix_expression_length);

    double result = parse_postfix_epxression(postfix_expression, postfix_expression_length);
    assert(fabs(2 - result) <= EPSILON);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        free(postfix_expression[i]);
    }
    free(postfix_expression);
}

void test_5() {
    // 3*(5-(7-1)*3)/2
    char *infix_expression[] = {"3", "*", "(", "5", "-", "(", "7", "-", "1", ")", "*", "3", ")", "/", "2"};
    int32_t infix_expression_length = 15;

    int32_t postfix_expression_length;
    char **postfix_expression = infix_to_postfix(infix_expression, infix_expression_length, &postfix_expression_length);

    double result = parse_postfix_epxression(postfix_expression, postfix_expression_length);
    assert(fabs(-19.5 - result) <= EPSILON);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        free(postfix_expression[i]);
    }
    free(postfix_expression);
}

void test_6() {
    // 5^(3/2)+(5)^(2/3)
    char *infix_expression[] = {"5", "^", "(", "3", "/", "2", ")", "+", "(", "5", ")", "^", "(", "2", "/", "3", ")"};
    int32_t infix_expression_length = 17;

    int32_t postfix_expression_length;
    char **postfix_expression = infix_to_postfix(infix_expression, infix_expression_length, &postfix_expression_length);

    double result = parse_postfix_epxression(postfix_expression, postfix_expression_length);
    assert(fabs(14.1043576257 - result) <= EPSILON);

    for (int32_t i = 0; i < postfix_expression_length; i += 1) {
        free(postfix_expression[i]);
    }
    free(postfix_expression);
}

int main() {
    printf("EPSILON = %.15lf\n", EPSILON);
    test_1();
    test_2();
    test_3();
    test_4();
    test_5();
    test_6();
}
